
/* ************************************************************************* */
/*
 *	compute_single_uuid.c
 *
 *	(C) 2010 VOCAL Technologies, Ltd.
 *
 *	This software is intended for use as a reference/standard implementation,
 *	and has been placed into the public domain by the copyright holder:
 *
 *	VOCAL Technologies, Ltd.
 *	90A John Muir Drive.
 *	Buffalo, NY  14228
 *
 *	Maintainer:  Dennis Towne
 */
/* ************************************************************************* */

#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "uuid_hash.h"
#include "util_tests.h"

#define dprintf(A) /* printf A */	/* debug printing macro */


/******************************************************************************/
/* uuid display/generate helpers */
/******************************************************************************/

void print_uuid_parms(struct test_data *td) {
	printf("patient record number %d\n", td->record_number);
	printf("patient_name          %s\n", td->patient_name);
	if (td->patient_ssn[0]) {
		printf("patient_ssn           %s\n", td->patient_ssn);
	}
	printf("patient_gender        %d\n", td->patient_gender);
	printf("patient_birth_year    %d\n", td->patient_year);
	printf("patient_birth_month   %d\n", td->patient_month);
	printf("patient_birth_day     %d\n", td->patient_day);

	if (td->policy_ssn[0]) {
		printf("policy_name           %s\n", td->policy_name);
		printf("policy_ssn            %s\n", td->policy_ssn);
		printf("policy_gender         %d\n", td->policy_gender);
		printf("policy_birth_year     %d\n", td->policy_year);
		printf("policy_birth_month    %d\n", td->policy_month);
		printf("policy_birth_day      %d\n", td->policy_day);
	}
	printf("patient_uuid          %s\n", td->uuid);
}

static void write_csv_header(FILE *outfile) {
	fprintf(outfile, "data_record,patient_name,patient_ssn,patient_gender,patient_birth_year,patient_birth_month,patient_birth_day,policy_name,policy_ssn,policy_gender,policy_birth_year,policy_birth_month,policy_birth_day,uuid\n");
}

#if 0 /* unused - using full line write instead */
static void write_csv_record(FILE *outfile, struct test_data *td) {
	fprintf(outfile, "%d,", td->record_number);
	fprintf(outfile, "%s,", td->patient_name);
	if (td->patient_ssn) {
		fprintf(outfile, "%d", td->patient_ssn);
	}
	fprintf(outfile, ",");
	if (td->patient_gender != 0) {
		fprintf(outfile, "%d", td->patient_gender);
	}
	fprintf(outfile, ",");
	if (td->patient_year != 0) {
		fprintf(outfile, "%d", td->patient_year);
	}
	fprintf(outfile, ",");
	if (td->patient_month != 0) {
		fprintf(outfile, "%d", td->patient_month);
	}
	fprintf(outfile, ",");
	if (td->patient_day != 0) {
		fprintf(outfile, "%d", td->patient_day);
	}
	fprintf(outfile, ",");

	fprintf(outfile, "%s,", td->policy_name);
	if (td->policy_ssn) {
		fprintf(outfile, "%d", td->policy_ssn);
	}
	fprintf(outfile, ",");

	if (td->policy_gender != 0) {
		fprintf(outfile, "%d", td->policy_gender);
	}
	fprintf(outfile, ",");
	if (td->policy_year != 0) {
		fprintf(outfile, "%d", td->policy_year);
	}
	fprintf(outfile, ",");
	if (td->policy_month != 0) {
		fprintf(outfile, "%d", td->policy_month);
	}
	fprintf(outfile, ",");
	if (td->policy_day != 0) {
		fprintf(outfile, "%d", td->policy_day);
	}
	fprintf(outfile, ",");

	fprintf(outfile, "%s", td->uuid);
	fprintf(outfile, "\n");
}
#endif

static void nuke_trailing_linefeed(char *read_buffer) {
	while(*read_buffer != 0 && *read_buffer != '\n' && *read_buffer != '\r') {
		read_buffer++;
	}
	*read_buffer = 0;
}


/******************************************************************************/
/* CSV record handling and extraction */
/******************************************************************************/

static const char *read_csv_string(const char *data, int *field_missing, char *dest, int dest_len) {
	if (data == NULL || *data == 0) {
		return NULL;
	}

	dprintf(("read csv string ->%s\n", data));
	*field_missing = 1;

	if (*data == '"') {
		int index = 0;
		data++;
		while(*data != 0 && *data != '"') {
			dest[index] = *data;
			if (dest_len > 1) {
				dest_len--;
				index++;
			}
			data++;
		}
		dest[index] = 0;
		if (dest[0] != 0) {
			*field_missing = 0;
		}
		if (*data == '"') {
			data++;
		}
	}
	else {
		int index = 0;
		while(*data != 0 && *data != ',') {
			dest[index] = *data;
			if (dest_len > 1) {
				dest_len--;
				index++;
			}
			data++;
		}
		dest[index] = 0;
		if (dest[0] != 0) {
			*field_missing = 0;
		}
	}

	/* advance to next field, if possible */
	if (*data == 0) {
		data = NULL;
	}
	else if (*data == ',') {
		data++;
	}

	return data;
}

static const char *read_csv_number(const char *data, int *field_missing, int *dest) {
	if (data == NULL || *data == 0) {
		return NULL;
	}

	dprintf(("read csv number ->%s\n", data));

	if (sscanf(data, "%d", dest) != 1) {
		*field_missing = 1;
	}
	else {
		*field_missing = 0;
	}

	/* advance to next field, if possible */
	if (*data == '-') {
		data++;
	}
	while(*data != 0 && *data != ',') {
		/* check for bogus chars in the field */
		if (*data < '0' || *data > '9') {
			*field_missing = 1;
			return NULL;
		}
		data++;
	}
	if (*data == 0) {
		data = NULL;
	}
	else if (*data == ',') {
		data++;
	}

	return data;
}


/******************************************************************************/
/* CSV record processing - read, parse, and generate uuid */
/******************************************************************************/

int process_csv_record(const char *data, struct test_data *td, int verbose) {
	char temp_buffer[1000];
	struct uuid_hash_data ud;
	int field_missing;

	/* Expected CSV format:
	 * data_record,patient_name,patient_ssn,patient_gender,patient_birth_year,patient_birth_month,patient_birth_day,policy_name,policy_ssn,policy_gender,policy_birth_year,policy_birth_month,policy_birth_day
	 * 2,Angelica,222446666,2,1910,1,1,Chloe,111222022,2,1916,12,2
	 */

	uuid_initialize(&ud);

	/*********************************************************/
	/* Read patient data fields */
	/*********************************************************/

	data = read_csv_string(data, &field_missing, td->patient_name, MAX_NAME_LEN);
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient name.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		uuid_hash_set_patient_name(&ud, td->patient_name);
	}

	/*-----------------------------*/
	data = read_csv_string(data, &field_missing, temp_buffer, sizeof(temp_buffer));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient ssn.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		/* SSN is handled as a string in the uuid_hash API */
	   	if (!uuid_hash_set_patient_ssn(&ud, temp_buffer)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad patient ssn.\n");
			}
			return 0;
		}
		strncpy(td->patient_ssn, temp_buffer, sizeof(td->patient_ssn));
		td->patient_ssn[MAX_SSN_LEN - 1] = 0;
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->patient_gender));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient gender.\n");
		}
		return 0;
	}
	else if (!field_missing) {
   		if (!uuid_hash_set_patient_gender(&ud, td->patient_gender)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad patient gender.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->patient_year));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient birth year.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_patient_birth_year(&ud, td->patient_year)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad patient birth year.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->patient_month));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient birth month.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_patient_birth_month(&ud, td->patient_month)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad patient birth month.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->patient_day));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at patient birth day.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_patient_birth_day(&ud, td->patient_day)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad patient birth day.\n");
			}
			return 0;
		}
	}


	/*********************************************************/
	/* Read policy holder data fields */
	/*********************************************************/

	data = read_csv_string(data, &field_missing, td->policy_name, MAX_NAME_LEN);
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at policy name.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		uuid_hash_set_policy_name(&ud, td->policy_name);
	}

	/*-----------------------------*/
	data = read_csv_string(data, &field_missing, temp_buffer, sizeof(temp_buffer));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at policy ssn.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		/* SSN is handled as a string in the uuid_hash API */
		if (!uuid_hash_set_policy_ssn(&ud, temp_buffer)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad policy ssn.\n");
			}
			return 0;
		}
		strncpy(td->policy_ssn, temp_buffer, sizeof(td->policy_ssn));
		td->policy_ssn[MAX_SSN_LEN - 1] = 0;
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->policy_gender));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at policy gender.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_policy_gender(&ud, td->policy_gender)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad policy gender.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->policy_year));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at policy birth year.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_policy_birth_year(&ud, td->policy_year)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad policy birth year.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->policy_month));
	if (data == NULL) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), short record at policy birth month.\n");
		}
		return 0;
	}
	else if (!field_missing) {
		if (!uuid_hash_set_policy_birth_month(&ud, td->policy_month)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad policy birth month.\n");
			}
			return 0;
		}
	}

	/*-----------------------------*/
	data = read_csv_number(data, &field_missing, &(td->policy_day));
	/* the last record doesn't have anything following it
	if (data == NULL) {
		printf("Error in process_csv_record(), short record at policy birth day.\n");
		return 0;
	}
	else
	*/
	if (!field_missing) {
		if (!uuid_hash_set_policy_birth_day(&ud, td->policy_day)) {
			if (verbose == 2) {
				printf("Error in process_csv_record(), bad policy birth day.\n");
			}
			return 0;
		}
	}


	/*********************************************************/
	/* Generate uuid, if possible */
	/*********************************************************/

	if (!uuid_create(&ud)) {
		if (verbose == 2) {
			printf("Error in process_csv_record(), uuid_create() failed.\n");
		}
		return 0;
	}

	strcpy(td->uuid, ud.uuid);
	return 1;
}

static int process_csv_file(const char *infile_name, const char *outfile_name, int verbose) {
	char read_buffer[1000];
	char record_id_buffer[1000];
	struct test_data td;
	FILE *infile;
	FILE *outfile;
	int counter = 0;
	int bad_counter = 0;
	int discard_counter = 0;

	infile = fopen(infile_name, "r");
	if (infile == NULL) {
		printf("Unable to open input file '%s'.\n", infile_name);
		return 1;
	}

	outfile = fopen(outfile_name, "w");
	if (outfile == NULL) {
		printf("Unable to open output file '%s'.\n", outfile_name);
		fclose(infile);
		return 1;
	}

	if (verbose) {
		printf("Reading patient records from input file '%s'.\n", infile_name);
		printf("Writing patient records to output file '%s'.\n", outfile_name);
		printf("\n");
	}

	/* strip header line */
	fgets(read_buffer, sizeof(read_buffer), infile);

	/* write new header */
	write_csv_header(outfile);

	while(1) {
		const char *data = read_buffer;
		int field_missing;

		if (fgets(read_buffer, sizeof(read_buffer), infile) == NULL) {
			if (verbose) {
				printf("\n");
			}
			break;
		}
		nuke_trailing_linefeed(read_buffer);
		dprintf(("Got line '%s'\n", read_buffer));

		init_test_data(&td);
		data = read_csv_string(data, &field_missing, record_id_buffer, sizeof(record_id_buffer));
		if (data == NULL || field_missing) {
			if (verbose) {
				printf("Ignoring line '%s'\n\n", read_buffer);
				printf("\n");
			}
			discard_counter++;

			/* ignore first header line */
			fprintf(outfile, "%s\n", read_buffer);
		}
		else if (!process_csv_record(data, &td, verbose)) {
			if (verbose) {
				printf("Unable to process patient data record '%s' from input file '%s'.\n", read_buffer, infile_name);
				printf("\n");
			}
			bad_counter++;

			/* ignore first header line */
			fprintf(outfile, "%s\n", read_buffer);
		}
		else {
			if (verbose) {
				printf("Processed patient data record '%s'.\n", record_id_buffer);
			}
			if (verbose == 2) {
				dprintf(("\n"));
				printf("Got record data:\n");
				print_uuid_parms(&td);
				printf("\n");
			}
			counter++;
			/* write_csv_record(outfile, &td); */
			fprintf(outfile, "%s,%s\n", read_buffer, td.uuid);
		}
	}

	if (verbose) {
		printf("\n");
		printf("Processed %d records, ignored %d bad records, discarded %d bad lines.\n", counter, bad_counter, discard_counter);
	}

	fclose(infile);
	fclose(outfile);
	return counter;
}


/******************************************************************************/
/* main routine and usage */
/******************************************************************************/

static void dump_test_usage(const char *name) {
	printf("Usage:\n");
	printf("%s <-v> <-vv> [input data filename] [output data filename]\n", name);
	printf("\n");
	printf("This program loads patient records from a file in CSV format, generates the\n");
	printf("UUID for each record, then writes the records including the UUID field to an\n");
	printf("output file.\n");
	printf("\n");
	printf("The '-v' and '-vv' options run the program in 'verbose' modes.\n");
	printf("This program exits 1 on failure, 0 on success.\n");
}

int main(int argc, char *argv[]) {
	const char *infile_name = NULL;
	const char *outfile_name = NULL;
	int verbose = 0;

	if (argc <= 1 || (argc == 2 && strcmp(argv[1], "-h") == 0)) {
		dump_test_usage(argv[0]);
		return 0;
	}

	if (strcmp(argv[1], "-v") == 0) {
		if (argc != 4) {
			dump_test_usage(argv[0]);
			return 0;
		}
		verbose = 1;
		infile_name  = argv[2];
		outfile_name = argv[3];
	}
	else if (strcmp(argv[1], "-vv") == 0) {
		if (argc != 4) {
			dump_test_usage(argv[0]);
			return 0;
		}
		verbose = 2;
		infile_name  = argv[2];
		outfile_name = argv[3];
	}
	else {
		if (argc != 3) {
			dump_test_usage(argv[0]);
			return 0;
		}
		infile_name  = argv[1];
		outfile_name = argv[2];
	}

	return process_csv_file(infile_name, outfile_name, verbose);
}

