
/* ************************************************************************* */
/*
 *	compute_single_uuid.c
 *
 *	(C) 2010 VOCAL Technologies, Ltd.
 *
 *	This software is intended for use as a reference/standard implementation,
 *	and has been placed into the public domain by the copyright holder:
 *
 *	VOCAL Technologies, Ltd.
 *	90A John Muir Drive.
 *	Buffalo, NY  14228
 *
 *	Maintainer:  Dennis Towne
 */
/* ************************************************************************* */

#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "uuid_hash.h"
#include "util_tests.h"

#define dprintf(A) /* printf A */	/* debug printing macro */


/******************************************************************************/
/* uuid display/generate helpers */
/******************************************************************************/

void print_uuid_parms(struct test_data *td) {
	printf("patient record number %d\n", td->record_number);
	printf("patient_name          %s\n", td->patient_name);
	if (td->patient_ssn[0]) {
		printf("patient_ssn           %s\n", td->patient_ssn);
	}
	printf("patient_gender        %d\n", td->patient_gender);
	printf("patient_birth_year    %d\n", td->patient_year);
	printf("patient_birth_month   %d\n", td->patient_month);
	printf("patient_birth_day     %d\n", td->patient_day);

	if (td->policy_ssn[0]) {
		printf("policy_name           %s\n", td->policy_name);
		printf("policy_ssn            %s\n", td->policy_ssn);
		printf("policy_gender         %d\n", td->policy_gender);
		printf("policy_birth_year     %d\n", td->policy_year);
		printf("policy_birth_month    %d\n", td->policy_month);
		printf("policy_birth_day      %d\n", td->policy_day);
	}
	printf("patient_uuid          %s\n", td->uuid);
}


/******************************************************************************/
/* demo routine - patient SSN available */
/******************************************************************************/

int read_patient_record(FILE *stream, struct test_data *td) {
	init_test_data(td);
	char temp[1000];

/* sample data record from the input file:
data_record           14
patient_name          Dominique
patient_gender        2
patient_birth_year    1934
patient_birth_month   11
patient_birth_day     26
policy_name           Crystal
policy_ssn            111222014
policy_gender         2
policy_birth_year     1910
policy_birth_month    1
policy_birth_day      1
data_end
*/

	while(fget_line(stream, temp, sizeof(temp))) {
		dprintf(("Got line '%s'\n", temp));

		char *next = next_arg(temp);
		if (*temp == 0 || *temp == '#') {
			/* do nothing - these are comment lines */
		}
		else if (wordcmp(temp, "data_end")) {
			return 1;
		}
		else if (next == NULL) {
			printf("Error in input file - got bad line '%s'.\n", temp);
			return 0;
		}

		else if (wordcmp(temp, "data_record")) {
			td->record_number = atoi(next);
		}

		else if (wordcmp(temp, "patient_name")) {
			strxcpy(td->patient_name, next, sizeof(td->patient_name));
		}
		else if (wordcmp(temp, "patient_ssn")) {
			strncpy(td->patient_ssn, next, sizeof(td->patient_ssn));
			td->patient_ssn[MAX_SSN_LEN - 1] = 0;
		}
		else if (wordcmp(temp, "patient_gender")) {
			td->patient_gender = atoi(next);
		}
		else if (wordcmp(temp, "patient_birth_year")) {
			td->patient_year = atoi(next);
		}
		else if (wordcmp(temp, "patient_birth_month")) {
			td->patient_month = atoi(next);
		}
		else if (wordcmp(temp, "patient_birth_day")) {
			td->patient_day = atoi(next);
		}

		else if (wordcmp(temp, "policy_name")) {
			strxcpy(td->policy_name, next, sizeof(td->patient_name));
		}
		else if (wordcmp(temp, "policy_ssn")) {
			strncpy(td->policy_ssn, next, sizeof(td->policy_ssn));
			td->policy_ssn[MAX_SSN_LEN - 1] = 0;
		}
		else if (wordcmp(temp, "policy_gender")) {
			td->policy_gender = atoi(next);
		}
		else if (wordcmp(temp, "policy_birth_year")) {
			td->policy_year = atoi(next);
		}
		else if (wordcmp(temp, "policy_birth_month")) {
			td->policy_month = atoi(next);
		}
		else if (wordcmp(temp, "policy_birth_day")) {
			td->policy_day = atoi(next);
		}
		else {
			printf("Error in input file - got bad line '%s'.\n", temp);
			return 0;
		}
	}

	return 0;
}

static int load_and_check_uuid(const char *filename, int quiet) {
	struct test_data td;
	FILE *stream;
	char temp[UUID_HASH_STRING_LENGTH + 1];

	stream = fopen(filename, "r");
	if (stream == NULL) {
		printf("Unable to open input file '%s'.\n", filename);
		return 1;
	}

	if (!quiet) {
		printf("Reading patient record from input file '%s'.\n", filename);
	}

	if (!read_patient_record(stream, &td)) {
		printf("Unable to read patient data record from input file '%s'.\n", filename);
		return 1;
	}

	if (!generate_test_uuid(&td, temp)) {
		printf("Unable to generate UUID - probably bad input data.\n");
		return 1;
	}

	if (!quiet) {
		printf("\n");
		printf("Read record data:\n");
		print_uuid_parms(&td);
	}
	printf("UUID: %s\n", temp);

	return 0;
}


/******************************************************************************/
/* main routine and usage */
/******************************************************************************/

static void dump_test_usage(const char *name) {
	printf("Usage:\n");
	printf("%s <-q> [input data filename]\n", name);
	printf("\n");
	printf("This program loads a single patient record from a file and generates the\n");
	printf("UUID for that record.  The file format is the same as that in the generated\n");
	printf("test vector file, plus the addition of a 'data_end' tag to mark the end\n");
	printf("of the record.\n");
	printf("\n");
	printf("The '-q' option runs the program in 'quiet' mode.\n");
	printf("This program exits 1 on failure, 0 on success.\n");
}

int main(int argc, char *argv[]) {
	const char *filename = NULL;
	int quiet = 0;

	if (argc <= 1 || (argc == 2 && strcmp(argv[1], "-h") == 0)) {
		dump_test_usage(argv[0]);
		return 0;
	}
	if (strcmp(argv[1], "-q") == 0) {
		if (argc == 2) {
			dump_test_usage(argv[0]);
			return 0;
		}
		quiet = 1;
		filename = argv[2];
	}
	else {
		filename = argv[1];
	}

	return load_and_check_uuid(filename, quiet);
}

