
/* ************************************************************************* */
/*
 *	uuid_hash.c
 *
 *	(C) 2010 VOCAL Technologies, Ltd.
 *
 *	This software is intended for use as a reference/standard implementation,
 *	and has been placed into the public domain by the copyright holder:
 *
 *	VOCAL Technologies, Ltd.
 *	90A John Muir Drive.
 *	Buffalo, NY  14228
 *
 *	Maintainer:  Dennis Towne
 */
/* ************************************************************************* */

#ifndef DO_OS_WINDOWS
#include <unistd.h>
#endif /* DO_OS_WINDOWS */
#include <stdio.h>
#include <string.h>

#include "system.h"
#include "util.h"
#include "sha2.h"
#include "uuid_hash.h"


#undef  DEBUG_DUMP
#define eprintf(A) printf A


#ifdef DEBUG_DUMP
/******************************************************************************/
/* debug/util */
/******************************************************************************/

void uuid_debug_dump_vars(struct uuid_hash_data *ud) {
	int i;

	printf("\n");
	printf("uuid_create() -----------------------------\n");
	printf("patient - name:  ");
	for(i = 0; i < UUID_FIRST_NAME_LEN; i++) {
		printf(" %2d", ud->patient_first_name[i]);
	}
	printf("\n");

	printf("patient - ssn:   ");
	for(i = 0; i < UUID_SSN_LEN; i++) {
		printf(" %2d", ud->patient_ssn[i]);
	}
	printf("\n");

	printf("patient - dob:  %4d\n",  ud->patient_day_of_birth);
	printf("patient - sex:  %4d\n",  ud->patient_gender);
	printf("patient - mob:  %4d\n",  ud->patient_month_of_birth);
	printf("patient - yob:  %4d\n",  ud->patient_year_of_birth);


	printf("\n");
	printf("\npolicy  - name:  ");
	for(i = 0; i < UUID_FIRST_NAME_LEN; i++) {
		printf(" %2d", ud->policy_first_name[i]);
	}
	printf("\n");

	printf("policy  - ssn:   ");
	for(i = 0; i < UUID_SSN_LEN; i++) {
		printf(" %2d", ud->policy_ssn[i]);
	}
	printf("\n");

	printf("policy  - dob:  %4d\n",  ud->policy_day_of_birth);
	printf("policy  - sex:  %4d\n",  ud->policy_gender);
	printf("policy  - mob:  %4d\n",  ud->policy_month_of_birth);
	printf("policy  - yob:  %4d\n",  ud->policy_year_of_birth);
	printf("flags:  %04x\n", ud->flags);
	printf("-------------------------------------------\n");
}

void byte_dump(uchar *data, int len) {
	int i;

	printf("byte_dump() - ");
	for(i = 0; i < len; i++) {
		printf(" %2d", data[i]);
	}
	printf("\n");
}
#endif /* DEBUG_DUMP */


/******************************************************************************/
/* conversion - ascii/ebcdic data to common format */
/******************************************************************************/

static int number_downconvert(int ch) {
	/* ebcdic has 0-9 mapped linearly, so this is acceptable */
	if (ch >= '0' && ch <= '9') {
		return ch - '0';
	}
	return -1;
}

static int alpha_downconvert(int ch) {
	/* ebcdic has alpha mapped in sections, this explicitly works everywhere */
	switch(ch) {
	case 'a':
	case 'A': return 1;

	case 'b':
	case 'B': return 2;

	case 'c':
	case 'C': return 3;

	case 'd':
	case 'D': return 4;

	case 'e':
	case 'E': return 5;

	case 'f':
	case 'F': return 6;

	case 'g':
	case 'G': return 7;

	case 'h':
	case 'H': return 8;

	case 'i':
	case 'I': return 9;

	case 'j':
	case 'J': return 10;

	case 'k':
	case 'K': return 11;

	case 'l':
	case 'L': return 12;

	case 'm':
	case 'M': return 13;

	case 'n':
	case 'N': return 14;

	case 'o':
	case 'O': return 15;

	case 'p':
	case 'P': return 16;

	case 'q':
	case 'Q': return 17;

	case 'r':
	case 'R': return 18;

	case 's':
	case 'S': return 19;

	case 't':
	case 'T': return 20;

	case 'u':
	case 'U': return 21;

	case 'v':
	case 'V': return 22;

	case 'w':
	case 'W': return 23;

	case 'x':
	case 'X': return 24;

	case 'y':
	case 'Y': return 25;

	case 'z':
	case 'Z': return 26;
	}

	return -1;
}

static int alphanum_upconvert(int ch) {
	/* ebcdic has alpha mapped in sections, this explicitly works everywhere */
	switch(ch) {
	case  0: return '0';
	case  1: return '1';
	case  2: return '2';
	case  3: return '3';
	case  4: return '4';

	case  5: return '5';
	case  6: return '6';
	case  7: return '7';
	case  8: return '8';
	case  9: return '9';

	case 10: return 'a';
	case 11: return 'b';
	case 12: return 'c';
	case 13: return 'd';
	case 14: return 'e';

	case 15: return 'f';
	case 16: return 'g';
	case 17: return 'h';
	case 18: return 'i';
	case 19: return 'j';

	case 20: return 'k';
	case 21: return 'l';
	case 22: return 'm';
	case 23: return 'n';
	case 24: return 'o';

	case 25: return 'p';
	case 26: return 'q';
	case 27: return 'r';
	case 28: return 's';
	case 29: return 't';

	case 30: return 'u';
	case 31: return 'v';
	}
	return -1;
}


/******************************************************************************/
/* clear/reset of the uuid structure */
/******************************************************************************/

void uuid_initialize(struct uuid_hash_data *ud) {
	ud->flags = 0;

	memset(ud->patient_first_name, 0, sizeof(ud->patient_first_name));
	memset(ud->patient_ssn, 0, sizeof(ud->patient_ssn));

	ud->patient_day_of_birth = 0;
	ud->patient_gender = 0;

	ud->patient_year_of_birth = 0;
	ud->patient_month_of_birth = 0;

	memset(ud->policy_first_name, 0, sizeof(ud->policy_first_name));
	memset(ud->policy_ssn, 0, sizeof(ud->policy_ssn));

	ud->policy_day_of_birth = 0;
	ud->policy_gender = 0;

	ud->policy_year_of_birth = 0;
	ud->policy_month_of_birth = 0;

	memset(ud->uuid, 0, sizeof(ud->uuid));
}


/******************************************************************************/
/* functions for properly filling out patient fields in the uuid data structure */
/******************************************************************************/

int uuid_hash_set_patient_name(struct uuid_hash_data *ud, const char *name) {
	int x;

	ud->flags |= UUID_FLAG_PATIENT_FN;
	if (name == NULL) {
		eprintf(("Error in uuid_hash_set_patient_name(), NULL name passed in.\n"));
		ud->flags -= UUID_FLAG_PATIENT_FN;
		return 0;
	}

	/* TODO - remove leading mr and mrs if present (?) possibly fail on any whitespace detection */
	x = 0;
	while(x < UUID_FIRST_NAME_LEN && *name != 0) {
		int temp = alpha_downconvert(*name);
		if (temp >= 0) {
			ud->patient_first_name[x] = temp;
			x++;
		}

		name++;
	}
	if (x == 0) {
		eprintf(("Error in uuid_hash_set_patient_name(), invalid name passed in.\n"));
		ud->flags -= UUID_FLAG_PATIENT_FN;
		return 0;
	}

	return 1;
}

int uuid_hash_set_patient_ssn(struct uuid_hash_data *ud, const char *ssn) {
	int x;

	ud->flags |= UUID_FLAG_PATIENT_SSN;
	if (ssn == NULL) {
		eprintf(("Error in uuid_hash_set_patient_ssn(), NULL ssn passed in.\n"));
		ud->flags -= UUID_FLAG_PATIENT_SSN;
		return 0;
	}

	x = 0;
	while(*ssn != 0) {
		int temp = number_downconvert(*ssn);
		if (temp >= 0) {
			if (x < UUID_SSN_LEN) {
				ud->patient_ssn[x] = temp;
			}
			x++;
		}

		ssn++;
	}
	if (x != UUID_SSN_LEN) {
		eprintf(("Error in uuid_hash_set_patient_ssn(), invalid ssn passed in.\n"));
		ud->flags -= UUID_FLAG_PATIENT_SSN;
		return 0;
	}
	/* TODO - stronger sanity checks on SSN */

	return 1;
}

int uuid_hash_set_patient_gender(struct uuid_hash_data *ud, int gender) {
	ud->flags |= UUID_FLAG_PATIENT_GENDER;
	if (gender != UUID_GENDER_MALE && gender != UUID_GENDER_FEMALE) {
		eprintf(("Error in uuid_hash_set_patient_gender(), bogus gender %d passed in.\n", gender));
		ud->flags -= UUID_FLAG_PATIENT_GENDER;
		return 0;
	}

	ud->patient_gender = gender;
	return 1;
}

int uuid_hash_set_patient_birth_year(struct uuid_hash_data *ud, int yob) {
	ud->flags |= UUID_FLAG_PATIENT_YOB;
	if (yob <= 1850 || yob >= 2100) {
		eprintf(("Error in uuid_hash_set_patient_birth_year(), bogus year %d passed in.\n", yob));
		ud->flags -= UUID_FLAG_PATIENT_YOB;
		return 0;
	}

	ud->patient_year_of_birth = yob;
	return 1;
}

int uuid_hash_set_patient_birth_month(struct uuid_hash_data *ud, int mob) {
	ud->flags |= UUID_FLAG_PATIENT_MOB;
	if (mob <= 0 || mob > 12) {
		eprintf(("Error in uuid_hash_set_patient_birth_month(), bogus month %d passed in.\n", mob));
		ud->flags -= UUID_FLAG_PATIENT_MOB;
		return 0;
	}

	ud->patient_month_of_birth = mob;
	return 1;
}

int uuid_hash_set_patient_birth_day(struct uuid_hash_data *ud, int dob) {
	ud->flags |= UUID_FLAG_PATIENT_DOB;
	if (dob <= 0 || dob > 31) {
		eprintf(("Error in uuid_hash_set_patient_birth_day(), bogus day %d passed in.\n", dob));
		ud->flags -= UUID_FLAG_PATIENT_DOB;
		return 0;
	}

	ud->patient_day_of_birth = dob;
	return 1;
}


/******************************************************************************/
/* functions for properly filling out policy holder fields in the uuid data structure */
/******************************************************************************/

int uuid_hash_set_policy_name(struct uuid_hash_data *ud, const char *name) {
	int x;

	ud->flags |= UUID_FLAG_POLICY_FN;
	if (name == NULL) {
		eprintf(("Error in uuid_hash_set_policy_name(), NULL name passed in.\n"));
		ud->flags -= UUID_FLAG_POLICY_FN;
		return 0;
	}

	/* TODO - remove leading mr and mrs if present (?) possibly fail on any whitespace detection */
	x = 0;
	while(x < UUID_FIRST_NAME_LEN && *name != 0) {
		int temp = alpha_downconvert(*name);
		if (temp >= 0) {
			ud->policy_first_name[x] = temp;
			x++;
		}

		name++;
	}
	if (x == 0) {
		eprintf(("Error in uuid_hash_set_policy_name(), invalid name passed in.\n"));
		ud->flags -= UUID_FLAG_POLICY_FN;
		return 0;
	}

	return 1;
}

int uuid_hash_set_policy_ssn(struct uuid_hash_data *ud, const char *ssn) {
	int x;

	ud->flags |= UUID_FLAG_POLICY_SSN;
	if (ssn == NULL) {
		eprintf(("Error in uuid_hash_set_policy_ssn(), NULL ssn passed in.\n"));
		ud->flags -= UUID_FLAG_POLICY_SSN;
		return 0;
	}

	x = 0;
	while(*ssn != 0) {
		int temp = number_downconvert(*ssn);
		if (temp >= 0) {
			if (x < UUID_SSN_LEN) {
				ud->policy_ssn[x] = temp;
			}
			x++;
		}

		ssn++;
	}
	if (x != UUID_SSN_LEN) {
		eprintf(("Error in uuid_hash_set_policy_ssn(), invalid ssn passed in.\n"));
		ud->flags -= UUID_FLAG_POLICY_SSN;
		return 0;
	}
	/* TODO - stronger sanity checks on SSN */

	return 1;
}

int uuid_hash_set_policy_gender(struct uuid_hash_data *ud, int gender) {
	ud->flags |= UUID_FLAG_POLICY_GENDER;
	if (gender != UUID_GENDER_MALE && gender != UUID_GENDER_FEMALE) {
		eprintf(("Error in uuid_hash_set_policy_gender(), bogus gender %d passed in.\n", gender));
		ud->flags -= UUID_FLAG_POLICY_GENDER;
		return 0;
	}

	ud->policy_gender = gender;
	return 1;
}

int uuid_hash_set_policy_birth_year(struct uuid_hash_data *ud, int yob) {
	ud->flags |= UUID_FLAG_POLICY_YOB;
	if (yob <= 1850 || yob >= 2100) {
		eprintf(("Error in uuid_hash_set_policy_birth_year(), bogus year %d passed in.\n", yob));
		ud->flags -= UUID_FLAG_POLICY_YOB;
		return 0;
	}

	ud->policy_year_of_birth = yob;
	return 1;
}

int uuid_hash_set_policy_birth_month(struct uuid_hash_data *ud, int mob) {
	ud->flags |= UUID_FLAG_POLICY_MOB;
	if (mob <= 0 || mob > 12) {
		eprintf(("Error in uuid_hash_set_policy_birth_month(), bogus month %d passed in.\n", mob));
		ud->flags -= UUID_FLAG_POLICY_MOB;
		return 0;
	}

	ud->policy_month_of_birth = mob;
	return 1;
}

int uuid_hash_set_policy_birth_day(struct uuid_hash_data *ud, int dob) {
	ud->flags |= UUID_FLAG_POLICY_DOB;
	if (dob <= 0 || dob > 31) {
		eprintf(("Error in uuid_hash_set_policy_birth_day(), bogus day %d passed in.\n", dob));
		ud->flags -= UUID_FLAG_POLICY_DOB;
		return 0;
	}

	ud->policy_day_of_birth = dob;
	return 1;
}


/******************************************************************************/
/* uuid calculations using SHA256 hash on pre-checked data structure */
/******************************************************************************/

static void uuid_patient_ssn_create(struct uuid_hash_data *ud) {
	SHA256state st;
	uchar  H1[SHA256dlen];
	uchar  H2[SHA256dlen];

	uchar  H1_buffer[UUID_H1_INPUT_LEN];
	uchar  H2_buffer[UUID_H2_INPUT_LEN];
	uchar *tp;
	int i;

#ifdef DEBUG_DUMP
	uuid_debug_dump_vars(ud);
#endif /* DEBUG_DUMP */

	/* first, calculate the H1 hash */
	tp = H1_buffer;
	memcpy(tp, ud->patient_first_name, UUID_FIRST_NAME_LEN);
	tp += UUID_FIRST_NAME_LEN;
	memcpy(tp, ud->patient_ssn, UUID_SSN_LEN);
	tp += UUID_SSN_LEN;
	*tp = ud->patient_day_of_birth;

	sha256init(&st);
	sha256(&st, H1_buffer, UUID_H1_INPUT_LEN);
	sha256finish(&st, H1);

#ifdef DEBUG_DUMP
	{ /* debug dump */
		char dump_buffer[2*SHA256dlen + 1];
		hex(H1, SHA256dlen, dump_buffer);
		printf("\n");
		byte_dump(H1_buffer, UUID_H1_INPUT_LEN);
		printf("uuid_create() - got H1 %s\n", dump_buffer);
	}
#endif /* DEBUG_DUMP */

	/* now calculate H2 */
	tp = H2_buffer;
	for(i = 0; i < UUID_H1_OUTPUT_NIBBLES; i++) {
		/* grab groups of four bits at a time */
		*tp = H1[i] & 15;
		tp++;
	}

	*tp = ud->patient_gender;
	tp++;
	*tp = ud->patient_month_of_birth;
	tp++;
	*tp = ud->patient_year_of_birth % 100;
	tp++;
	*tp = ud->patient_year_of_birth / 100;

	sha256init(&st);
	sha256(&st, H2_buffer, UUID_H2_INPUT_LEN);
	sha256finish(&st, H2);

#ifdef DEBUG_DUMP
	{ /* debug dump */
		char dump_buffer[2*SHA256dlen + 1];
		hex(H2, SHA256dlen, dump_buffer);
		printf("\n");
		byte_dump(H2_buffer, UUID_H2_INPUT_LEN);
		printf("uuid_create() - got H2 %s\n", dump_buffer);
	}
#endif /* DEBUG_DUMP */

	/* create final uuid, five bits at a time */

	/* make sure reserved 0x10 bit in first byte is set */
	ud->uuid[0] = alphanum_upconvert((H2[0] & 0x0f) | 0x10);

	/* convert remaining bytes */
	for(i = 1; i < UUID_HASH_STRING_LENGTH; i++) {
		ud->uuid[i] = alphanum_upconvert(H2[i] & 31);
	}

	/* terminate the string */
	ud->uuid[i] = 0;
}


static void uuid_patient_nossn_create(struct uuid_hash_data *ud) {
	SHA256state st;
	uchar  H3[SHA256dlen];

	uchar  H3_buffer[UUID_H3_MAX_INPUT_LEN];
	uchar *tp;
	int i;

#ifdef DEBUG_DUMP
	uuid_debug_dump_vars(ud);
#endif /* DEBUG_DUMP */

	sha256init(&st);

	/* --------------------------------------*/
	/* construct the first block for hashing */
	/* --------------------------------------*/

	tp = H3_buffer;
	memcpy(tp, ud->policy_first_name, UUID_FIRST_NAME_LEN);
	tp += UUID_FIRST_NAME_LEN;
	memcpy(tp, ud->policy_ssn, UUID_SSN_LEN);
	tp += UUID_SSN_LEN;

	*tp = ud->policy_gender;
	tp++;
	*tp = ud->policy_day_of_birth;
	tp++;
	*tp = ud->policy_month_of_birth;
	tp++;
	*tp = ud->policy_year_of_birth % 100;
	tp++;
	*tp = ud->policy_year_of_birth / 100;

	sha256(&st, H3_buffer, UUID_H3_BLOCK_1_LEN);

#ifdef DEBUG_DUMP
	{ /* debug dump */
		printf("\n");
		byte_dump(H3_buffer, UUID_H3_BLOCK_1_LEN);
	}
#endif /* DEBUG_DUMP */

	/* --------------------------------------*/
	/* construct the first block for hashing */
	/* --------------------------------------*/

	tp = H3_buffer;
	memcpy(tp, ud->patient_first_name, UUID_FIRST_NAME_LEN);
	tp += UUID_FIRST_NAME_LEN;

	*tp = ud->patient_gender;
	tp++;
	*tp = ud->patient_day_of_birth;
	tp++;
	*tp = ud->patient_month_of_birth;
	tp++;
	*tp = ud->patient_year_of_birth % 100;
	tp++;
	*tp = ud->patient_year_of_birth / 100;

	sha256(&st, H3_buffer, UUID_H3_BLOCK_2_LEN);

#ifdef DEBUG_DUMP
	{ /* debug dump */
		printf("\n");
		byte_dump(H3_buffer, UUID_H3_BLOCK_1_LEN);
	}
#endif /* DEBUG_DUMP */

	/* --------------------------------------*/
	/* finish the hash and do final formatting */
	/* --------------------------------------*/

	sha256finish(&st, H3);

#ifdef DEBUG_DUMP
	{ /* debug dump */
		char dump_buffer[2*SHA256dlen + 1];
		hex(H3, SHA256dlen, dump_buffer);
		printf("uuid_create() - got H3 %s\n", dump_buffer);
	}
#endif /* DEBUG_DUMP */

	/* create final uuid, five bits at a time */

	/* mask out reserved 0x10 bit in first byte */
	ud->uuid[0] = alphanum_upconvert(H3[0] & 0x0f);

	/* convert remaining bytes */
	for(i = 1; i < UUID_HASH_STRING_LENGTH; i++) {
		ud->uuid[i] = alphanum_upconvert(H3[i] & 0x1f);
	}

	/* terminate the string */
	ud->uuid[i] = 0;
}


/******************************************************************************/
/* router function - figure out which uuid hash to use based on patient SSN */
/******************************************************************************/

int uuid_create(struct uuid_hash_data *ud) {
	int i;

	/* sanity checks */
	if (ud == NULL) {
		return 0;
	}

	/* construct an all 0 string as a default */
	for(i = 0; i < UUID_HASH_STRING_LENGTH; i++) {
		ud->uuid[i] = '0';
	}
	ud->uuid[i] = 0;

	/* check various success cases */
	if ((ud->flags & UUID_PATIENT_FLAGS_OK) == UUID_PATIENT_FLAGS_OK) {
		/* we have enough information to create a direct patient UUID */
		uuid_patient_ssn_create(ud);
		return 1;
	}
	else if (ud->flags == UUID_PATIENT_NOSSN_FLAGS_OK) {
		uuid_patient_nossn_create(ud);
		return 1;
	}
	else if ((ud->flags & UUID_PATIENT_SSN_FLAGS) == 0) {
		/* the default return is all zeros, which is what we should return here */
		return 1;
	}

	eprintf(("Error in uuid_create(), structure not properly filled out.  Flags %04x.\n", ud->flags));
	return 0;
}


